''Mofified by Dave Ratcliff 8/6/12 to show Igx, Igy and Igz gyro variables to serial terminal - control output to motors disabled.
'' will need to give a little throttle from transmitter to see IgX variables start to accumulate.
'' Igx = positive #'s leaning craft towards it's left
'' Igy = positive #'s leaning craft towards it's front
'' Igz = positive #'s twisting the C.C.W.
''
''
''
''  QuadRotor - Full Stability
''     -- By Jason Dorie --               
''               
''
'' This code assumes:
'' - an R/C receiver connected to pins 0 - 3
'' - an ITG-3200 with pullups connected to pins 16 & 17
'' - four electronic speed controls (ESCs) connected to pins 8 - 11
'' - a serial interface (PropPlug) on pins 30 & 31
''
'' Other configurations are certainly possible, but will require changes to the code
''
'' Most pin assignments can be easily altered in the Main function below, with the exception
'' of the ESCs - The Servo8Fast code has been altered to support only a single bank of 8 outputs
'' at very high speed.  The current bank of 8 is hard-coded to P8-P15, however it can be changed
'' by adjusting which segments are commented out near the top of the DAT section.    




CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

  '' Motor pins for '+' shaped frame config
  OUT_F  = 0                    ' Front ESC
  OUT_R  = 1                    ' Right ESC
  OUT_B  = 2                    ' Back ESC
  OUT_L  = 3                    ' Left ESC

  '' Motor pins for 'X' shaped frame config
  OUT_FL = 0                    ' Front Left ESC                       
  OUT_FR = 1                    ' Front Right ESC
  OUT_BR = 2                    ' Back Right ESC                        
  OUT_BL = 3                    ' Back Left ESC


  '' Indices for R/C receiver inputs 
  IN_THRO = 0
  IN_AILE = 1
  IN_ELEV = 2
  IN_RUDD = 3
  IN_GAIN = 4
  IN_AUX  = 5
  IN_PING = 6


  Filter_Heavy =    $2AAAAAAA         '0.1666 ** 2^32  (RC = 0.0200)   
  Filter_Moderate = $2F9DB22D         '0.1860 ** 2^32  (RC = 0.0175)   
  Filter_Light =    $4923A29C         '0.2857 ** 2^32  (RC = 0.0100)   



OBJ
  Gyro  : "ITG-3200-pasm.spin"                          ' 1 cog
  RC    : "RC_Receiver_7_opt.spin"                      ' 1 cog                       
  ESC   : "Servo8Fast.spin"                             ' 1 cog
  Dbg   : "FullDuplexSerial"
  
  RollPID   : "IntPID-scaled.spin"
  PitchPID  : "IntPID-scaled.spin"
  YawPID    : "IntPID-scaled.spin"

  'Eeprom    : "Basic_I2C_Driver.spin"
      
VAR
  long Output[4]                                        ' Temp output flight controls array, copied into 'servo' when finished                        
  long counter
  long Thro, Aile, Elev, Rudd, Gain, Ping
  long iAile, iElev, iRudd
  long iGx, iGy, iGz                                    ' Gyro readings (integrated)    
  
  long DesiredRoll, DesiredPitch, DesiredYaw
  
  long ThroMix
  long LastTime, CurTime
  long Roll, Pitch, Yaw
  long GXLP, GYLP, GZLP
  long DoIntegrate

  long Gx, Gy, Gz                                       ' Gyro readings (instant)
  
  long DataMode                                         ' What data to send to the host                        
  long rxResult
         
  long RollOut, PitchOut, YawOut
PUB Main
    
  'Give the compass time to intialize
  'waitcnt( 80 * 200 + cnt )
 
  Dbg.Start( 31, 30, 0, 115200 )
  Gyro.Start( 17, 16 )
  DataMode := 0                 'Send adjusted gyro readings


  ' Initialize servo values to all neutral (zero for throttles)
  ' Delay values of 1500 = center, 1000 = min, 2000 = max

  if( ThrottleStartupTime == 0 )
    Output[0] := 1000    
    Output[1] := 1000                                   ' motors off    
    Output[2] := 1000    
    Output[3] := 1000
  else
    Output[0] := 2000    
    Output[1] := 2000                     ' motors to full throttle on startup    
    Output[2] := 2000    
    Output[3] := 2000


  'Start the servo driver cog
  ESC.AddPin(8)
  ESC.AddPin(9)
  ESC.AddPin(10)
  ESC.AddPin(11)  
  ESC.Set(  8, Output[0] )
  ESC.Set(  9, Output[1] )
  ESC.Set( 10, Output[2] )
  ESC.Set( 11, Output[3] )
  ESC.Start


  'ProgramLoop


  'Motor startup max throttle setting, if specified
  counter := (ThrottleStartupTime * 80_000 )            'Convert value to milliseconds
  if( counter => 400 )
    waitcnt( counter + cnt )

    Output[0] := 1000    
    Output[1] := 1000                                   ' motors off    
    Output[2] := 1000    
    Output[3] := 1000
    
    ESC.Set(  8, Output[0] )
    ESC.Set(  9, Output[1] )
    ESC.Set( 10, Output[2] )
    ESC.Set( 11, Output[3] )
     

  'Initialize the RC input object
  RC.setpins( %111_1111 )           'Input pins are P0 - P6
  RC.start  

  'These numbers are scaled by 2^32 - the PID code internally divides the result by 2^32 on output
  'The shifts in the Init calls are just to allow the use of more human-readable values.

  RollPID.Init(    PitchRoll_Pgain,  PitchRoll_Igain,  PitchRoll_Dgain )
  PitchPID.Init(   PitchRoll_Pgain,  PitchRoll_Igain,  PitchRoll_Dgain )
  YawPID.Init(     Yaw_PGain,        Yaw_IGain,        Yaw_DGain )                    

  RollPID.SetD2Gain(  PitchRoll_D2Gain )
  PitchPID.SetD2Gain( PitchRoll_D2Gain )
  'YawPID.SetD2Gain(   Yaw_D2Gain )


  'Should probably wait for the gyro to finish startup here, but it takes a while
  'for the ESCs to initialize anyway...

  'repeat until( Gyro.IsReady == 1 )

  FlightLoop



PUB FlightLoop | t

  counter := 0
  LastTime := cnt

  Roll := Pitch := Yaw := 0
  DesiredRoll := DesiredPitch := DesiredYaw := 0
  Gx := Gy := Gz := 0

  repeat
         
    {rxResult := Dbg.rxCheck
    if( rxResult <> -1 )
      if( rxResult == $AA )
        DoConfig
        LastTime := cnt           'Reset the loop wait counter or we'll be waiting a while
       
      elseif( rxResult == $AB )
        DataMode := 0
        
      elseif( rxResult == $AC )
        DataMode := 1
    '}
  
    Thro := RC.getrc(IN_THRO)
    Aile := RC.getrc(IN_AILE)
    Elev := RC.getrc(IN_ELEV)
    Rudd := RC.getrc(IN_RUDD)
    Gain := RC.getrc(IN_GAIN)
    'Ping := RC.get(IN_PING) / 147


    'Narrow deadband around zero on inputs
    if( ||Aile < 5 )
      Aile := 0
    if( ||Elev < 5 )
      Elev := 0
    if( ||Rudd < 5 )
      Rudd := 0


    iAile += Aile
    iElev += Elev
    iRudd += Rudd


    if( GyroOrientation == 0 )
      GXLP := -Gyro.GetRX          ' Gyro is mounted with X = left/right, Y = front/back
      GYLP := Gyro.GetRY
      GZLP := Gyro.GetRZ
    elseif( GyroOrientation == 1 )
      GYLP := Gyro.GetRX           ' Gyro is mounted with X = front/back, Y = left/right (90-degrees counter clockwise)
      GXLP := Gyro.GetRY
      GZLP := Gyro.GetRZ
    else '( GyroOrientation == 2 )
      GYLP := -Gyro.GetRX          ' Gyro is mounted with X = front/back, Y = left/right (90-degrees clockwise)
      GXLP := -Gyro.GetRY
      GZLP := Gyro.GetRZ


    if( Gain > 0 )                '(gear switch up - light filter)
      Gx += Filter_Light ** (GXLP - Gx) 
      Gy += Filter_Light ** (GYLP - Gy)
      Gz += Filter_Light ** (GZLP - Gz)
    else                          '(gear switch down - no filter)
      Gx := GXLP
      Gy := GYLP
      Gz := GZLP
           
    {elseif( Gain > 100 )
      'Low pass filtered numbers   (gear up, aux2 mid)
      Gx += Filter_Moderate ** (GXLP - Gx) 
      Gy += Filter_Moderate ** (GYLP - Gy)
      Gz += Filter_Moderate ** (GZLP - Gz)
    elseif( Gain > -300 )
      'Low pass filtered numbers   (gear up, aux2 down)
      Gx += Filter_Heavy ** (GXLP - Gx) 
      Gy += Filter_Heavy ** (GYLP - Gy)
      Gz += Filter_Heavy ** (GZLP - Gz)
    else
      'Both switches toward me - filter off
      Gx := GXLP 
      Gy := GYLP
      Gz := GZLP }
          
    'I might need to include scale factors for gyro rate here.  Mine is 14.375 lsb's per/dec/sec
    'with a maximum rate of 2000 deg/sec.  Others may need to be scaled up / down to fall into the
    'same control range.  Using (GVal << 8) ** ScaleFactor should allow enough range...  maybe?
       
    iGx += Gx
    iGy += Gy
    iGz += Gz
         
    'Zero all targets when throttle is off - makes for more stable liftoff
    if( Thro < -370 )
      iGx := iAile := 0
      iGy := iElev := 0
      iGz := iRudd := 0
      DoIntegrate := false
    else
      DoIntegrate := true
             
    DesiredRoll := iAile * PitchRollControlScale 
    DesiredPitch := iElev * PitchRollControlScale
    DesiredYaw := iRudd * YawControlScale
             
    RollOut := RollPID.Calculate( DesiredRoll , iGx , DoIntegrate )    
    PitchOut := PitchPID.Calculate( DesiredPitch , iGy , DoIntegrate )    
    YawOut := YawPID.Calculate( DesiredYaw , iGz , DoIntegrate )    
             
    ThroMix := (Thro + 400) ~> 2                        ' Approx 0 - 256
    ThroMix <#= 64                                      ' Above 1/4 throttle, clamp it to 64
    ThroMix #>= 0

    'add 1500 to all Output values to make them 'servo friendly' again
    Thro += 1500
        
    if( FrameConfiguration == 0 )
      ' + configuration
      Output[OUT_F] := Thro + ((-PitchOut + YawOut) * ThroMix) ~> 6                          
      Output[OUT_R] := Thro + (( RollOut  - YawOut) * ThroMix) ~> 6  
      Output[OUT_B] := Thro + (( PitchOut + YawOut) * ThroMix) ~> 6
      Output[OUT_L] := Thro + ((-RollOut  - YawOut) * ThroMix) ~> 6
       
    else
    ' X configuration
      Output[OUT_FL] := Thro + ((-PitchOut - RollOut - YawOut) * ThroMix) ~> 6                          
      Output[OUT_FR] := Thro + ((-PitchOut + RollOut + YawOut) * ThroMix) ~> 6  
      Output[OUT_BL] := Thro + (( PitchOut - RollOut + YawOut) * ThroMix) ~> 6
      Output[OUT_BR] := Thro + (( PitchOut + RollOut - YawOut) * ThroMix) ~> 6
            
    Output[0] := 1000 #> Output[0] <# 2000
    Output[1] := 1000 #> Output[1] <# 2000
    Output[2] := 1000 #> Output[2] <# 2000
    Output[3] := 1000 #> Output[3] <# 2000
              
    'Copy new Ouput array into servo values
    {ESC.Set(  8, Output[0] )
    ESC.Set(  9, Output[1] )
    ESC.Set( 10, Output[2] )
    ESC.Set( 11, Output[3] ) }
    Dbg.tx(1)
    Dbg.str(string("Igx = ",11))
    Dbg.dec(igx)
    Dbg.str(string(13,13,"Igy = ",11))
    Dbg.dec(igy)
    Dbg.str(string(13,13,"Igz = ",11))
    Dbg.dec(igz)
    Dbg.tx(13)
    waitcnt(clkfreq/100+cnt)    
    {waitcnt( constant(80_000_000 / 250) + LastTime )
    LastTime += constant(80_000_000 / 250)}
       

DAT

ConfigStart

GyroOrientation         long  2                 ' 0 = normal, 1 = rotated 90 CCW, 2 = rotated 90 CW
FrameConfiguration      long  1                 ' 0 = +, 1 = x 

PitchRollControlScale   long  6                 ' 16 (insane), 8 (rapid), 4 (agile), 2 (slower), 1 (beginner)  (or anywhere in between)
YawControlScale         long  6

PitchRoll_PGain         long  $10_0000           'jason's pid's for the Elev-8
PitchRoll_IGain         long  $0  
PitchRoll_DGain         long  $1500_0000
PitchRoll_D2Gain        long  0   '$1000_0000

Yaw_PGain               long  $28_0000
Yaw_IGain               long  $0  
Yaw_DGain               long  $1800_0000
       
ThrottleStartupTime     long  0                 'Hold time in mSec (0 = no throttle range set on startup)
DummyVar                long  0,0,0,0           'Padding, to keep the EEPROM write at 64 bytes

' The previous data is 16 longs, or 64 bytes total
ConfigEnd